#include "DisplayRGB.h"


#ifdef __AVR__
#include <avr/pgmspace.h>
#include <pgmspace.h>
#endif
#include "glcdfont.c"

//#ifdef __Petitfatfs__
#include <petit_fatfs.h>
#include <pff.h>
#include <integer.h>
#include <diskio.h>
//#endif


//#define pgm_read_byte(addr) (*(const unsigned char *)(addr))

#ifndef pgm_read_byte
#define pgm_read_byte(addr) (*(const unsigned char *)(addr))
#endif
#ifndef pgm_read_word
#define pgm_read_word(addr) (*(const unsigned short *)(addr))
#endif
#ifndef pgm_read_dword
#define pgm_read_dword(addr) (*(const unsigned long *)(addr))
#endif

// Pointers are a peculiar case...typically 16-bit on AVR boards,
// 32 bits elsewhere.  Try to accommodate both...

#if !defined(__INT_MAX__) || (__INT_MAX__ > 0xFFFF)
#define pgm_read_pointer(addr) ((void *)pgm_read_dword(addr))
#else
#define pgm_read_pointer(addr) ((void *)pgm_read_word(addr))
#endif

#ifndef min
#define min(a,b) (((a) < (b)) ? (a) : (b))
#endif

// the memory buffer for the LCD
uint8_t _displayBuffer[1024];

//#ifdef __Petitfatfs__
// drawBitmapFromSd() return values
#define BMP_OK       0 // ok
#define BMP_NO_FILE  1 // error: bitmap file not found on SD
#define BMP_NOT_WBM  2 // error: not a Windows Bitmap
#define BMP_NOT_1BPP 3 // error: not a monochrome bitmap
#define BMP_ERR_CMP  4 // error: compressed bitmap
#define BMP_TOO_BIG  5 // error: bitmap width > Gamebuino screen

// Various macros
#define WORD(AR,PO) (byte)AR[PO]+0x100L*(byte)AR[PO+1] // read word from char array AR at position PO
#define LONG(AR,PO) (byte)AR[PO]+0x100L*(byte)AR[PO+1]+0x10000L*(byte)AR[PO+2]+0x1000000L*(byte)AR[PO+3]

// Following two functions are needed by petit_fatfs
byte rx() { return SPI.transfer(0xff); }
void tx(byte d) { SPI.transfer(d); }

// Read 'n' bytes of bitmap data starting at position 'start' and put them in buffer buf[]
void readNCharsAt(byte n, word start, char *buf) {
//#ifdef debug // use PROGMEM dump of bitmap data
	//for (byte i = 0; i<n; i++) { buf[i] = pgm_read_byte(&(bmp[start + i])); }
//#else        // open real bitmap on SD
	WORD br; pf_lseek(start); pf_read((void*)&buf[0], n, &br);
//#endif
}

void Display::tftBegin(uint8_t cs,uint8_t wr ,uint8_t rs , uint8_t rest ,uint8_t pmw)
{
	this->_cs = cs;
	//this->_rd = rd;
	this->_wr = wr;
	this->_rs = rs;
	this->_rest = rest;
	this->_pmw = pmw;

	//cursorY = cursorX = 0;
	//fontSize = 1;
	//color = BLACK;
	//bgcolor = WHITE;
	textWrap = true;
	
	pinMode(this->_cs,OUTPUT);
	//pinMode(this->_rd,OUTPUT);
	pinMode(this->_wr,OUTPUT);
	pinMode(this->_rs,OUTPUT);
	pinMode(this->_rest,OUTPUT);
	pinMode(this->_pmw,OUTPUT);
	
	
	this->csport = portOutputRegister(digitalPinToPort(this->_cs));
	//this->pmwport = portOutputRegister(digitalPinToPort(this->_pmw));
	this->wrport = portOutputRegister(digitalPinToPort(this->_wr));
	this->rsport = portOutputRegister(digitalPinToPort(this->_rs));
	this->restport = portOutputRegister(digitalPinToPort(this->_rest));
	
	
	this->cspinmask  = digitalPinToBitMask(this->_cs);
	//this->pmwpinmask  = digitalPinToBitMask(this->_pmw);
	this->wrpinmask  = digitalPinToBitMask(this->_wr);
	this->rspinmask  = digitalPinToBitMask(this->_rs);
	this->restpinmask  = digitalPinToBitMask(this->_rest);
	
	*(this->csport) |= this->cspinmask;
	//*(this->pmwport) |= this->pmwpinmask;
	*(this->wrport) |= this->wrpinmask;
	*(this->rsport) |= this->rspinmask;
	*(this->restport) |= this->restpinmask;		
}


/*#if 0 //defined(ARDUINO_ARCH_SAM)
	void Display::write_delay()
	{	
		volatile uint32_t delaycnt = 10;
		while( delaycnt-- );
	}
#endif*/


void Display::write_spi(const uint8_t data)
{
	uint8_t bit,i;
	for(bit = 0x01,i = 0 ; bit ; i++,bit <<= 1)	
	{
		if(data & bit) *(this->dataport[i]) |= this->datapinmask[i];
    else   *(this->dataport[i]) &= (~this->datapinmask[i]);
    	
    //write_delay();
  }
}

void Display::setdatapin(uint8_t b0,uint8_t b1, uint8_t b2 , uint8_t b3,
													uint8_t b4, uint8_t b5, uint8_t b6, uint8_t b7)
{
	uint8_t i;
	this->data[0] = b0;
	this->data[1] = b1;
	this->data[2] = b2;
	this->data[3] = b3;
	this->data[4] = b4;
	this->data[5] = b5;
	this->data[6] = b6;
	this->data[7] = b7;
	for(i = 0; i < 8 ;i++)
		pinMode(this->data[i],OUTPUT);
	for(i = 0; i < 8 ;i++)
		dataport[i] = portOutputRegister(digitalPinToPort(this->data[i]));
	for(i = 0; i < 8 ;i++)
		datapinmask[i] = digitalPinToBitMask(this->data[i]);
	
	this->setbacklight(255);
}

void Display::setbacklight(uint8_t pmw)
{
	analogWrite(this->_pmw, pmw);
}


void Display::write_wr_reg(uint16_t data)
{
 	
 	*(this->rsport) &= (~this->rspinmask);
 	write_delay();
 	*(this->csport) &= (~this->cspinmask);
 	write_delay();
 	this->write_spi( data >> 8 );
 	write_delay();
 	*(this->wrport) &= (~this->wrpinmask);
 	write_delay(); 	
	*(this->wrport) |= this->wrpinmask;
 	write_delay();
 	this->write_spi( (data&0x00ff) );
 	write_delay();
 	*(this->wrport) &= (~this->wrpinmask);
 	write_delay();
	*(this->wrport) |= this->wrpinmask;
	write_delay();
 	*(this->csport) |= this->cspinmask;
 	write_delay();
}

void Display::write_wr_data(uint16_t data)
{
	*(this->rsport) |= this->rspinmask;
	write_delay();
	*(this->csport) &= (~this->cspinmask);
	write_delay();
	
	this->write_spi( data >> 8 );
	write_delay();
	*(this->wrport) &= (~this->wrpinmask);
	write_delay();
	*(this->wrport) |= this->wrpinmask;
	write_delay();
	
	this->write_spi( (data&0x00ff) );
	write_delay();
	*(this->wrport) &= (~this->wrpinmask);
	write_delay();
	*(this->wrport) |= this->wrpinmask;
	write_delay();
	
	*(this->csport) |= this->cspinmask;
	write_delay();
}

void Display::write_cmd(uint16_t cmd ,uint16_t data)
{
	this->write_wr_reg(cmd);
	this->write_wr_data(data);
}

void Display::set_window(uint8_t x1,uint8_t y1,uint8_t x2,uint8_t y2)
{
	this->write_wr_reg(0x0037);this->write_wr_data(x1); // windows address 
	this->write_wr_reg(0x0036);this->write_wr_data(x2);
	this->write_wr_reg(0x0039);this->write_wr_data(y1);    
	this->write_wr_reg(0x0038);this->write_wr_data(y2);	
	this->write_wr_reg(0x0020);  //HS
	this->write_wr_data(x1); 		 
	this->write_wr_reg(0x0021);  //VS
	this->write_wr_data(y1);
}

void Display::clear(void) {
	memset(_displayBuffer, 0, LCDWIDTH * LCDHEIGHT / 8);
	cursorY = cursorX = 0;
}

void Display::begin()
{
		if( this->_rest )
		{
			 	*(this->restport) &= (~this->restpinmask);					
			 	delay(50); //     Delay 50ms
			 	*(this->restport) |= this->restpinmask;		
			 	delay(10); //     Delay 10ms
		}
	
		this->write_cmd(0x0001, 0x011C); // set SS and NL bit
    this->write_cmd(0x0002, 0x0100); // set 1 line inversion
    this->write_cmd(0x0003, 0x1030); // set GRAM write direction and BGR=1,??16BIT 65K?MDT1=0,MDT0=0
    this->write_cmd(0x0008, 0x0808); // set BP and FP
    this->write_cmd(0x000C, 0x0000); // RGB interface setting R0Ch=0x0110 for RGB 18Bit and R0Ch=0111for

    this->write_cmd(0x000F, 0x0e01); // Set frame rate
   
//*************Power On sequence ****************//
    delay(50); //     Delay 50ms
    this->write_cmd(0x0010, 0x0A00); // Set SAP,DSTB,STB
    this->write_cmd(0x0011, 0x1038); // Set APON,PON,AON,VCI1EN,VC
    //delay(50); //     Delay 50ms
    this->write_cmd(0x00ff, 0x0003);    //
    this->write_cmd(0x00b0, 0x1411);    //Set VCOM 1d
    this->write_cmd(0x00b1, 0x0202);    //GVCL/GVDD voltage setting
    this->write_cmd(0x00b2, 0x0313);  //VCL voltage setting

//------------------------ Set GRAM area  set window --------------------------------//
    this->write_cmd (0x0030, 0x0000);//???????
    this->write_cmd (0x0031, 0x00db);
    this->write_cmd (0x0032, 0x0000);
    this->write_cmd (0x0033, 0x0000);//??????
    this->write_cmd (0x0034, 0x00db);
    this->write_cmd (0x0035, 0x0000);//R34H,R35H???????
    this->write_cmd (0x0036, 0x00AF);
    this->write_cmd (0x0037, 0x0000);
    this->write_cmd (0x0038, 0x00DB);
    this->write_cmd (0x0039, 0x0000);
    delay(20);
   
    this->write_cmd(0x00ff, 0x0003);
//    WriteRegister(0x00b0, 0x1d01);
// ----------- Adjust the Gamma Curve ----------//
     this->write_cmd(0x0050, 0x0000);
     this->write_cmd(0x0051, 0x0300);
     this->write_cmd(0x0052, 0x0103);
     this->write_cmd(0x0053, 0x2011);
     this->write_cmd(0x0054, 0x0703);
     this->write_cmd(0x0055, 0x0000);
     this->write_cmd(0x0056, 0x0400);
     this->write_cmd(0x0057, 0x0107);
     this->write_cmd(0x0058, 0x2011);
     this->write_cmd(0x0059, 0x0703);
     delay(50); //     Delay 50ms
     this->write_cmd(0x0020, 0x0000); // Set GRAM Address
     this->write_cmd(0x0021, 0x0000); // Set GRAM Address
     this->write_cmd(0x0007, 0x1017);
	
		
		this->fillRect(0,0,LCDWIDTH,LCDHEIGHT,0xffff);
}

// Bresenham's algorithm - thx wikpedia
void Display::drawLine(int16_t x0, int16_t y0,
	int16_t x1, int16_t y1,
	uint16_t color)
{
	int16_t steep = abs(y1 - y0) > abs(x1 - x0);
	if (steep) {
		swap(x0, y0);
		swap(x1, y1);
	}

	if (x0 > x1) {
		swap(x0, x1);
		swap(y0, y1);
	}

	int16_t dx, dy;
	dx = x1 - x0;
	dy = abs(y1 - y0);

	int16_t err = dx / 2;
	int16_t ystep;

	if (y0 < y1) {
		ystep = 1;
	}
	else {
		ystep = -1;
	}

	for (; x0 <= x1; x0++) {
		if (steep) {
			drawPixel(y0, x0, color);
		}
		else {
			drawPixel(x0, y0, color);
		}
		err -= dy;
		if (err < 0) {
			y0 += ystep;
			err += dx;
		}
	}
}

void Display::drawFastVLine(int16_t x, int16_t y,
	int16_t h, uint16_t color)
{
	// Update in subclasses if desired!
	drawLine(x, y, x, y + h - 1, color);
}


void Display::drawFastHLine(int16_t x, int16_t y,
	int16_t w, uint16_t color)
{
	// Update in subclasses if desired!
	drawLine(x, y, x + w - 1, y, color);
}

// Draw a rectangle
void Display::drawRect(int16_t x, int16_t y,
	int16_t w, int16_t h,
	uint16_t color)
{
	drawFastHLine(x, y, w, color);
	drawFastHLine(x, y + h - 1, w, color);
	drawFastVLine(x, y, h, color);
	drawFastVLine(x + w - 1, y, h, color);
}

void Display::fillRect(int16_t x, int16_t y, int16_t w, int16_t h,
	uint16_t color)
{
	// Update in subclasses if desired!
	for (int16_t i = x; i<x + w; i++) {
		drawFastVLine(i, y, h, color);
	}
}

void Display::fillScreen(uint16_t color) {
	fillRect(0, 0, _width, _height, color);
}

void Display::invertDisplay(boolean i)
{
	// Do nothing, must be subclassed if supported
}

// draw a circle outline
void Display::drawCircle(int16_t x0, int16_t y0, int16_t r,
	uint16_t color) {
	int16_t f = 1 - r;
	int16_t ddF_x = 1;
	int16_t ddF_y = -2 * r;
	int16_t x = 0;
	int16_t y = r;

	drawPixel(x0, y0 + r, color);
	drawPixel(x0, y0 - r, color);
	drawPixel(x0 + r, y0, color);
	drawPixel(x0 - r, y0, color);

	while (x<y) {
		if (f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x;

		drawPixel(x0 + x, y0 + y, color);
		drawPixel(x0 - x, y0 + y, color);
		drawPixel(x0 + x, y0 - y, color);
		drawPixel(x0 - x, y0 - y, color);
		drawPixel(x0 + y, y0 + x, color);
		drawPixel(x0 - y, y0 + x, color);
		drawPixel(x0 + y, y0 - x, color);
		drawPixel(x0 - y, y0 - x, color);
	}
}

void Display::drawCircleHelper(int16_t x0, int16_t y0,
	int16_t r, uint8_t cornername, uint16_t color) {
	int16_t f = 1 - r;
	int16_t ddF_x = 1;
	int16_t ddF_y = -2 * r;
	int16_t x = 0;
	int16_t y = r;

	while (x<y) {
		if (f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x;
		if (cornername & 0x4) {
			drawPixel(x0 + x, y0 + y, color);
			drawPixel(x0 + y, y0 + x, color);
		}
		if (cornername & 0x2) {
			drawPixel(x0 + x, y0 - y, color);
			drawPixel(x0 + y, y0 - x, color);
		}
		if (cornername & 0x8) {
			drawPixel(x0 - y, y0 + x, color);
			drawPixel(x0 - x, y0 + y, color);
		}
		if (cornername & 0x1) {
			drawPixel(x0 - y, y0 - x, color);
			drawPixel(x0 - x, y0 - y, color);
		}
	}
}

void Display::fillCircle(int16_t x0, int16_t y0, int16_t r,
	uint16_t color)
{
	drawFastVLine(x0, y0 - r, 2 * r + 1, color);
	fillCircleHelper(x0, y0, r, 3, 0, color);
}

// used to do circles and roundrects
void Display::fillCircleHelper(int16_t x0, int16_t y0, int16_t r,
	uint8_t cornername, int16_t delta, uint16_t color)
{
	int16_t f = 1 - r;
	int16_t ddF_x = 1;
	int16_t ddF_y = -2 * r;
	int16_t x = 0;
	int16_t y = r;

	while (x<y) {
		if (f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x;

		if (cornername & 0x1) {
			drawFastVLine(x0 + x, y0 - y, 2 * y + 1 + delta, color);
			drawFastVLine(x0 + y, y0 - x, 2 * x + 1 + delta, color);
		}
		if (cornername & 0x2) {
			drawFastVLine(x0 - x, y0 - y, 2 * y + 1 + delta, color);
			drawFastVLine(x0 - y, y0 - x, 2 * x + 1 + delta, color);
		}
	}
}

// Draw a triangle
void Display::drawTriangle(int16_t x0, int16_t y0,
	int16_t x1, int16_t y1,
	int16_t x2, int16_t y2, uint16_t color) {
	drawLine(x0, y0, x1, y1, color);
	drawLine(x1, y1, x2, y2, color);
	drawLine(x2, y2, x0, y0, color);
}

// Fill a triangle
void Display::fillTriangle(int16_t x0, int16_t y0,
	int16_t x1, int16_t y1,
	int16_t x2, int16_t y2, uint16_t color) {
	int16_t a, b, y, last;

	// Sort coordinates by Y order (y2 >= y1 >= y0)
	if (y0 > y1) {
		swap(y0, y1); swap(x0, x1);
	}
	if (y1 > y2) {
		swap(y2, y1); swap(x2, x1);
	}
	if (y0 > y1) {
		swap(y0, y1); swap(x0, x1);
	}

	if (y0 == y2) { // Handle awkward all-on-same-line case as its own thing
		a = b = x0;
		if (x1 < a)      a = x1;
		else if (x1 > b) b = x1;
		if (x2 < a)      a = x2;
		else if (x2 > b) b = x2;
		drawFastHLine(a, y0, b - a + 1, color);
		return;
	}

	int16_t
		dx01 = x1 - x0,
		dy01 = y1 - y0,
		dx02 = x2 - x0,
		dy02 = y2 - y0,
		dx12 = x2 - x1,
		dy12 = y2 - y1,
		sa = 0,
		sb = 0;

	// For upper part of triangle, find scanline crossings for segments
	// 0-1 and 0-2.  If y1=y2 (flat-bottomed triangle), the scanline y1
	// is included here (and second loop will be skipped, avoiding a /0
	// error there), otherwise scanline y1 is skipped here and handled
	// in the second loop...which also avoids a /0 error here if y0=y1
	// (flat-topped triangle).
	if (y1 == y2) last = y1;   // Include y1 scanline
	else         last = y1 - 1; // Skip it

	for (y = y0; y <= last; y++) {
		a = x0 + sa / dy01;
		b = x0 + sb / dy02;
		sa += dx01;
		sb += dx02;
		/* longhand:
		a = x0 + (x1 - x0) * (y - y0) / (y1 - y0);
		b = x0 + (x2 - x0) * (y - y0) / (y2 - y0);
		*/
		if (a > b) swap(a, b);
		drawFastHLine(a, y, b - a + 1, color);
	}

	// For lower part of triangle, find scanline crossings for segments
	// 0-2 and 1-2.  This loop is skipped if y1=y2.
	sa = dx12 * (y - y1);
	sb = dx02 * (y - y0);
	for (; y <= y2; y++) {
		a = x1 + sa / dy12;
		b = x0 + sb / dy02;
		sa += dx12;
		sb += dx02;
		/* longhand:
		a = x1 + (x2 - x1) * (y - y1) / (y2 - y1);
		b = x0 + (x2 - x0) * (y - y0) / (y2 - y0);
		*/
		if (a > b) swap(a, b);
		drawFastHLine(a, y, b - a + 1, color);
	}
}

// Draw a rounded rectangle
void Display::drawRoundRect(int16_t x, int16_t y, int16_t w,
	int16_t h, int16_t r, uint16_t color)
{
	// smarter version
	drawFastHLine(x + r, y, w - 2 * r, color); // Top
	drawFastHLine(x + r, y + h - 1, w - 2 * r, color); // Bottom
	drawFastVLine(x, y + r, h - 2 * r, color); // Left
	drawFastVLine(x + w - 1, y + r, h - 2 * r, color); // Right
													   // draw four corners
	drawCircleHelper(x + r, y + r, r, 1, color);
	drawCircleHelper(x + w - r - 1, y + r, r, 2, color);
	drawCircleHelper(x + w - r - 1, y + h - r - 1, r, 4, color);
	drawCircleHelper(x + r, y + h - r - 1, r, 8, color);
}

// Fill a rounded rectangle
void Display::fillRoundRect(int16_t x, int16_t y, int16_t w,
	int16_t h, int16_t r, uint16_t color)
{
	// smarter version
	fillRect(x + r, y, w - 2 * r, h, color);

	// draw four corners
	fillCircleHelper(x + w - r - 1, y + r, r, 1, h - 2 * r - 1, color);
	fillCircleHelper(x + r, y + r, r, 2, h - 2 * r - 1, color);
}

//Adafruit GFX
// Draw a 1-bit image (bitmap) at the specified (x,y) position from the
// provided bitmap buffer (must be PROGMEM memory) using the specified
// foreground color (unset bits are transparent).
void Display::drawBitmap1(int16_t x, int16_t y, const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color) {

  int16_t i, j, byteWidth = (w + 7) / 8;
  uint8_t byte;

  for(j=0; j<h; j++) {
    for(i=0; i<w; i++) {
      if(i & 7) byte <<= 1;
      else      byte   = pgm_read_byte(bitmap + j * byteWidth + i / 8);
      if(byte & 0x80) drawPixel(x+i, y+j, color);

    }
  }
}

//Adafruit GFX
// Draw a 1-bit image (bitmap) at the specified (x,y) position from the
// provided bitmap buffer (must be PROGMEM memory) using the specified
// foreground (for set bits) and background (for clear bits) colors.
void Display::drawBitmap2(int16_t x, int16_t y,
 const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, uint16_t bg) {

  int16_t i, j, byteWidth = (w + 7) / 8;
  uint8_t byte;

  for(j=0; j<h; j++) {
    for(i=0; i<w; i++ ) {
      if(i & 7) byte <<= 1;
      else      byte   = pgm_read_byte(bitmap + j * byteWidth + i / 8);
      if(byte & 0x80) drawPixel(x+i, y+j, color);
      else            drawPixel(x+i, y+j, bg);
    }
  }
}

//Adafruit GFX
// drawBitmap() variant for RAM-resident (not PROGMEM) bitmaps.
void Display::drawBitmap3(int16_t x, int16_t y,
 uint8_t *bitmap, int16_t w, int16_t h, uint16_t color) {

  int16_t i, j, byteWidth = (w + 7) / 8;
  uint8_t byte;

  for(j=0; j<h; j++) {
    for(i=0; i<w; i++ ) {
      if(i & 7) byte <<= 1;
      else      byte   = bitmap[j * byteWidth + i / 8];
      if(byte & 0x80) drawPixel(x+i, y+j, color);
    }
  }
}

//Adafruit GFX
// drawBitmap() variant w/background for RAM-resident (not PROGMEM) bitmaps.
void Display::drawBitmap4(int16_t x, int16_t y,
 uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, uint16_t bg) {

  int16_t i, j, byteWidth = (w + 7) / 8;
  uint8_t byte;

  for(j=0; j<h; j++) {
    for(i=0; i<w; i++ ) {
      if(i & 7) byte <<= 1;
      else      byte   = bitmap[j * byteWidth + i / 8];
      if(byte & 0x80) drawPixel(x+i, y+j, color);
      else            drawPixel(x+i, y+j, bg);
    }
  }
}

//Adafruit GFX
//Draw XBitMap Files (*.xbm), exported from GIMP,
//Usage: Export from GIMP to *.xbm, rename *.xbm to *.c and open in editor.
//C Array can be directly used with this function
void Display::drawXBitmap(int16_t x, int16_t y,
 const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color) {

  int16_t i, j, byteWidth = (w + 7) / 8;
  uint8_t byte;

  for(j=0; j<h; j++) {
    for(i=0; i<w; i++ ) {
      if(i & 7) byte >>= 1;
      else      byte   = pgm_read_byte(bitmap + j * byteWidth + i / 8);
      if(byte & 0x01) drawPixel(x+i, y+j, color);
    }
  }
}

//GAMEBUINO
void Display::drawBitmap(int8_t x, int8_t y, int8_t w, int8_t h, const uint8_t *bitmap, uint8_t dx, uint8_t dy, uint8_t dw, uint8_t dh, uint16_t color) {
	int8_t i, j, byteWidth = (w + 7) / 8;
	dw += dx;
	dh += dy;
	int8_t largest = 0;
	int8_t largesty = 0;
	for (j = 0; j < h; j++) {
		for (i = 0; i < w; i++) {
			if (pgm_read_byte(bitmap + j * byteWidth + i / 8) & (B10000000 >> (i % 8))) {
				int8_t drawX = x + i;
				int8_t drawY = y + j;

				if (drawX >= dx && drawX < dw && drawY >= dy && drawY < dh) {
					drawPixel(x, y, color);
				}
			}
		}
	}
}

// Draw inverted-color bitmap from SRAM char array
void drawBitmapA(byte x, byte y, char *bitmap, uint16_t color) { // where bitmap[] is the char array containing bitmap data
	byte w = bitmap[0];
	byte h = bitmap[1];
	byte i, j, byteNum, bitNum, byteWidth = (w + 7) >> 3;
	for (i = 0; i < w; i++) {
		byteNum = i / 8; bitNum = i % 8;
		for (j = 0; j < h; j++) {
			if (!((bitmap[2 + j*byteWidth + byteNum]) & (B10000000 >> bitNum))) {
				drawPixel(x + i, y + j, color);
			}
		}
	}
}

// Draw inverted-color bitmaps from SRAM for use with tilemap duhjoker
void drawBitmapB(byte x, byte y, char *bitmap, uint8_t dx, uint8_t dy, uint8_t dw, uint8_t dh, uint16_t color) { // where bitmap[] is the char array containing bitmap data
	byte w = bitmap[0];
	byte h = bitmap[1];
	byte i, j, byteNum, bitNum, byteWidth = (w + 7) >> 3;
	dw += dx;
	dh += dy;
	for (i = 0; i<w; i++) {
		byteNum = i / 8; bitNum = i % 8;
		for (j = 0; j<h; j++) {
			if (!((bitmap[2 + j*byteWidth + byteNum]) & (B10000000 >> bitNum))) {
				int8_t drawX = x + i;
				int8_t drawY = y + j;

				if (drawX >= dx && drawX < dw && drawY >= dy && drawY < dh) {
					drawPixel(x, y, color);
				}
			}
		}
	}
}

// Draw bitmap from SD at position x0, y0
int drawBitmapFromSd(byte x0, byte y0, char *bitmap, uint16_t color) { // where bitmap[] is the name of the bitmap file on SD
	char buf[13]; // max 2 bytes for W/H + 11 bytes of data to cover the 84 pixel width of the  Gamebuino screen
	// 1. open bitmap file
    // open real bitmap on SD
	(pf_open(bitmap)); return BMP_NO_FILE;
	// 2a. read information from bitmap header > identifier
	readNCharsAt(2, 0, buf); // BMP.wIdentifier
	buf[2] = NULL;
	if (strcmp(buf, "BM")) return BMP_NOT_WBM;
	// 2b. read information from bitmap header > bpp, compression and data size
	readNCharsAt(10, 28, buf); // BMP.wBitsPerPixel + BMP.dwCompression + BMP.dwDataSize
	if (WORD(buf, 0) != 1) return BMP_NOT_1BPP; // not a 1bpp bitmap
	if (LONG(buf, 2)) return BMP_ERR_CMP; // not an UNcompressed bitmap
	word datasize = WORD(buf, 6);
	// 2c. read information from bitmap header > data offset
	readNCharsAt(4, 10, buf); // BMP.dwDataOffset
	word offset = WORD(buf, 0);
	// 2d. read information from bitmap header > dimensions
	readNCharsAt(8, 18, buf); // BMP.dwWidth + BMP.dwHeight
	byte w = (byte)buf[0];
	if (w>LCDWIDTH) return BMP_TOO_BIG;
	byte h = (byte)buf[4];
	byte datalen = datasize / h;
	// 3. read and draw bitmap data
	buf[0] = w; buf[1] = 1; // get bitmap data line by line
	for (byte y = 0; y<h; y++) {
		readNCharsAt(datalen, offset + y*datalen, buf + 2);
		drawBitmapA(x0, y0 + (h - 1) - y, buf, color); // and draw it!
	}
	return BMP_OK;
}

//byte slide;


void Display::drawTilemap(int x, int y, const uint8_t *tilemap, const uint8_t **spritesheet, uint16_t color) {
	drawTilemap(x, y, tilemap, spritesheet, 0, 0, LCDWIDTH, LCDHEIGHT, color);
}

void Display::drawTilemap(int x, int y, const uint8_t *tilemap, const uint8_t **spritesheet, uint8_t dx, uint8_t dy, uint8_t dw, uint8_t dh, uint16_t color) {
	uint8_t tilemap_width = pgm_read_byte(tilemap);
	uint8_t tilemap_height = pgm_read_byte(tilemap + 1);
	uint8_t tile_width = pgm_read_byte(tilemap + 2);
	uint8_t tile_height = pgm_read_byte(tilemap + 3);
	tilemap += 4; // now the first tiyleis at tilemap
	uint8_t ddw = dw + dx;
	uint8_t ddh = dh + dy;
	uint8_t maxDdx = (dw - x + tile_width - 1) / tile_width;
	uint8_t maxDdy = (dh - y + tile_height - 1) / tile_height;
	if (tilemap_width < maxDdx) {
		maxDdx = tilemap_width;
	}
	if (tilemap_height < maxDdy) {
		maxDdy = tilemap_height;
	}
	int8_t startDdx = (-x) / tile_width;
	int8_t startDdy = (-y) / tile_height;
	if (startDdx < 0) {
		startDdx = 0;
	}
	if (startDdy < 0) {
		startDdy = 0;
	}
	if (flagcollision)numcolision = 0;                                 //Line 735 - clear numcolision - ADD by Summoner123

	for (uint8_t ddy = startDdy; ddy < maxDdy; ddy++) {
		for (uint8_t ddx = startDdx; ddx < maxDdx; ddx++) {
			int8_t drawX = ddx*tile_width + x + dx;
			int8_t drawY = ddy*tile_height + y + dy;
			uint8_t tile = pgm_read_byte(tilemap + ddy*tilemap_width + ddx);
			if (drawX >= dx && drawY >= dy && drawX <= (ddw - tile_width) && drawY <= (ddh - tile_height)) {
				drawBitmap1(drawX, drawY, spritesheet[tile], tile_width, tile_height, color);

				if (flagcollision) {
					solid[numcolision].x = drawX;                     //Save X coordinate      - ADD by Summoner123
					solid[numcolision].y = drawY;                     //Save Y coordinate      - ADD by Summoner123
					solid[numcolision].spritecol = spritesheet[tile]; //Save Sprite of tile    - ADD by Summoner123
					numcolision++;                                    //Increment numcolision  - ADD by Summoner123
				}
			}
			else { // we need to draw a partial bitmap
				drawBitmap(drawX, drawY, tile_width, tile_height, spritesheet[tile], dx, dy, dw, dh, color);
			}
		}
	}
}




boolean Display::getBitmapPixel(const uint8_t* bitmap, uint8_t x, uint8_t y) {
	return pgm_read_byte(bitmap + 2 + y * ((pgm_read_byte(bitmap) + 7) / 8) + (x >> 3)) & (B10000000 >> (x % 8));
}

#if ARDUINO >= 100
size_t Display::write(uint8_t c) {
#else
void Display::write(uint8_t c) {
#endif

	if (c == '\n') {
		cursorY += fontSize * fontHeight;
		cursorX = 0;
	}
	else if (c == '\r') {
		// skip em
	}
	else {
		drawChar(cursorX, cursorY, c, WHITE, BLACK, fontSize);
		cursorX += fontSize * fontWidth;
		if (textWrap && (cursorX > (LCDWIDTH - fontSize * fontWidth))) {
			cursorY += fontSize * fontHeight;
			cursorX = 0;
		}
	}
#if ARDUINO >= 100
	return 1;
#endif
}

void Display::setFont(const uint8_t *f) {
	font = (uint8_t*)f;
	fontWidth = pgm_read_byte(font) + 1;
	fontHeight = pgm_read_byte(font + 1) + 1;
	font += 2; //offset the pointer to star after the width and height bytes
}

	// draw a character
	void Display::drawChar(int16_t x, int16_t y, unsigned char c,
		uint16_t color, uint16_t bg, uint8_t size)
	{
		if ((x >= _width) || // Clip right
			(y >= _height) || // Clip bottom
			((x + 6 * size - 1) < 0) || // Clip left
			((y + 8 * size - 1) < 0))     // Clip top
			return;

		for (int8_t i = 0; i < 6; i++) {
			uint8_t line;
			if (i == 5)
				line = 0x0;
			else
				line = pgm_read_byte(font + (c * 5) + i);
			for (int8_t j = 0; j < 8; j++) {
				if (line & 0x1) {
					if (size == 1) // default size
						drawPixel(x + i, y + j, color);
					else {  // big size
						fillRect(x + (i*size), y + (j*size), size, size, color);
					}
				}
				else if (bg != color) {
					if (size == 1) // default size
						drawPixel(x + i, y + j, bg);
					else {  // big size
						fillRect(x + i*size, y + j*size, size, size, bg);
					}
				}
				line >>= 1;
			}
		}
	}



	void Display::setCursor(int16_t x, int16_t y)
	{
		cursor_x = x;
		cursor_y = y;
	}


	void Display::setTextSize(uint8_t s)
	{
		textsize = (s > 0) ? s : 1;
	}



	void Display::setTextColor(uint16_t c) {
		textcolor = c;
		textbgcolor = c;
		// for 'transparent' background, we'll set the bg 
		// to the same as fg instead of using a flag
	}

	void Display::setTextColor(uint16_t c, uint16_t b)
	{
		textcolor = c;
		textbgcolor = b;
	}

	void Display::setTextWrap(boolean w)
	{
		wrap = w;
	}

	void Display::setColor(uint16_t color) {
		color = color;
		color = color;
		// for 'transparent' background, we'll set the bg 
		// to the same as fg instead of using a flag
	}

//	void Display::setColor(uint16_t color, uint16_t color)
//	{
//		color = color;
//		color = color;
//	}

	uint8_t Display::getRotation(void)
	{
		return rotation;
	}

	void Display::setRotation(uint8_t x)
	{
		rotation = (x & 3);
		switch (x) {
		case 0:
		case 2:
			_width = WIDTH;
			_height = HEIGHT;
			break;
		case 1:
		case 3:
			_width = HEIGHT;
			_height = WIDTH;
			break;
		}
	}

	// return the size of the display (per current rotation)
	int16_t Display::width(void)
	{
		return _width;
	}

	int16_t Display::height(void)
	{
		return _height;
	}

	/***************************************************************/
	uint16_t RGB(uint16_t r, uint16_t g, uint16_t b)
	{
		uint16_t rgb = 0;
		r *= 31;
		r /= 255;

		g *= 63;
		g /= 255;

		b *= 31;
		b /= 255;

		rgb |= (r << 11);
		rgb |= (g << 6);
		rgb |= b;
		return rgb;
	}


